@@ -1489,7 +1489,7 @@ int ha_myisam::assign_to_keycache(THD* thd, HA_CHECK_OPT *check_opt)
     param->db_name=    table->s->db.str;
     param->table_name= table->s->table_name.str;
     param->testflag= 0;
-    mi_check_print_error(param, errmsg);
+    mi_check_print_error(param, "%s", errmsg);
   }
   DBUG_RETURN(error);
 }
@@ -1555,7 +1555,7 @@ int ha_myisam::preload_keys(THD* thd, HA_CHECK_OPT *check_opt)
     param->db_name=    table->s->db.str;
     param->table_name= table->s->table_name.str;
     param->testflag=   0;
-    mi_check_print_error(param, errmsg);
+    mi_check_print_error(param, "%s", errmsg);
     DBUG_RETURN(error);
   }
 }
